from yams.buildobjs import (
    EntityType,
    RelationType,
    SubjectRelation,
    Float,
    Date,
    String,
    RichString,
)
from cubicweb.schema import WorkflowableEntityType, RRQLExpression, ERQLExpression
from cubicweb.schema import RQLVocabularyConstraint

from cubicweb import _
from cubicweb.schemas.base import CWUser

CWUser.add_relation(
    String(maxsize=32, description=_("social security number")), name="ssnum"
)
CWUser.add_relation(
    SubjectRelation("PostalAddress", cardinality="??", composite="subject"),
    name="lives_at",
)


class Expense(WorkflowableEntityType):
    __permissions__ = {
        "read": ("users", "managers"),
        "add": ("users", "managers"),
        "update": ("managers", ERQLExpression('X in_state S, NOT S name "accepted"')),
        "delete": ("managers", ERQLExpression('X in_state S, NOT S name "accepted"')),
    }
    title = String(maxsize=128, required=True, fulltextindexed=True)
    description = RichString(fulltextindexed=True)

    # XXX '*' cardinality since ExpendLine has to be a line of an Expense and
    # may also be a line of a Refund
    has_lines = SubjectRelation("ExpenseLine", cardinality="+*", composite="subject")
    # workflow : submitted, accepted


class ExpenseLine(EntityType):
    __permissions__ = {
        "read": ("users", "managers"),
        "add": ("users", "managers"),
        "update": (
            "managers",
            ERQLExpression(
                'E is Expense, E has_lines X, E in_state S, NOT S name "accepted"'
            ),
        ),
        "delete": (
            "managers",
            ERQLExpression(
                'E is Expense, E has_lines X, E in_state S, NOT S name "accepted"'
            ),
        ),
    }
    title = String(maxsize=256, required=True, fulltextindexed=True)
    diem = Date(required=True)
    type = String(
        required=True,
        internationalizable=True,
        vocabulary=(
            _("transport"),
            _("accomodation"),
            _("food"),
            _("communication"),
            _("others"),
        ),
    )
    amount = Float(required=True, description=_("total amount including taxes"))
    taxes = Float(required=True, description=_("total tax"))
    taxes_currency = String(required=True, maxsize=10, default="EUR")
    taxes_exchange_rate = Float(required=True, default=1.0)

    paid_by = SubjectRelation(
        "PaidByAccount",
        cardinality="1*",
        constraints=[
            RQLVocabularyConstraint(
                'NOT (O associated_to A, A in_state ST, ST name "deactivated")'
            ),
            RQLVocabularyConstraint('NOT (O account "")'),
        ],
    )
    paid_for = SubjectRelation(
        "PaidForAccount",
        cardinality="+*",
        constraints=[
            RQLVocabularyConstraint('NOT O account ""'),
        ],
    )

    currency = String(required=True, maxsize=30, default="EUR")
    exchange_rate = Float(required=True, default=1.0)


class PaidByAccount(EntityType):
    __permissions__ = {
        "read": ("users", "managers"),
        "add": ("managers",),
        "update": ("managers",),
        "delete": ("managers",),
    }
    label = String(required=True, maxsize=128)
    account = String(maxsize=16)

    associated_to = SubjectRelation("CWUser", cardinality="??")


class PaidForAccount(EntityType):
    __permissions__ = {
        "read": ("users", "managers"),
        "add": ("managers",),
        "update": ("managers",),
        "delete": ("managers",),
    }
    label = String(required=True, maxsize=128)
    account = String(maxsize=16)


class Refund(WorkflowableEntityType):
    __permissions__ = {
        "read": ("users", "managers"),
        "add": ("managers",),
        "update": ("managers",),
        "delete": ("managers",),
    }
    payment_date = Date(description=_("payment date"))
    payment_mode = String(maxsize=64, description=_("payment mode"))

    # XXX '*' cardinality since ExpendLine has to be a line of an Expense and
    # may also be a line of a Refund
    has_lines = SubjectRelation("ExpenseLine", cardinality="+*")
    to_account = SubjectRelation("PaidByAccount", cardinality="1*")
    # workflow : preparation / paid


class has_lines(RelationType):
    # note: The RRQLExpression used will only work when adding ExpenseLines on
    # Expense, and not on refund. ExpenseLines are supposed to be added to
    # Refund automatically via a hook through an unsafe_execute
    __permissions__ = {
        "read": ("managers", "users"),
        "add": (
            "managers",
            RRQLExpression('S is Expense, S in_state ST, NOT ST name "accepted"'),
        ),
        "delete": (
            "managers",
            RRQLExpression('S is Expense, S in_state ST, NOT ST name "accepted"'),
        ),
    }


class has_attachment(RelationType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", RRQLExpression("U has_update_permission S", "S")),
        "delete": ("managers", RRQLExpression("U has_update_permission S", "S")),
    }
    subject = ("Expense", "ExpenseLine")
    object = "File"



change_relation_props("Expense", "title", "String", fulltextindexed=True)
change_relation_props("ExpenseLine", "title", "String", fulltextindexed=True)

add_cube("file")
add_relation_definition("Expense", "has_attachment", "File")
add_relation_definition("ExpenseLine", "has_attachment", "File")



"""specific views for expense component

:organization: Logilab
:copyright: 2008-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

import os
from io import BytesIO

from logilab.mtconverter import xml_escape
from logilab.common.registry import yes

from cubicweb.predicates import (
    one_line_rset,
    is_instance,
    rql_condition,
    authenticated_user,
)
from cubicweb_web.view import EntityView
from cubicweb_web import action, component
from cubicweb_web.views import (
    uicfg,
    primary,
    autoform,
    workflow,
    urlrewrite,
    ibreadcrumbs,
    tableview,
)

from cubicweb import _


class ExpenseLineIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("IBreadCrumbs")

    def parent_entity(self):
        return self.entity.parent_expense


class ExpenseURLRewriter(urlrewrite.SimpleReqRewriter):
    rules = [
        (
            "/todo",
            dict(rql="Any E,S WHERE E is Expense, " 'E in_state S, S name "submitted"'),
        ),
    ]


_abaa = uicfg.actionbox_appearsin_addmenu

_abaa.tag_subject_of(("*", "has_attachment", "*"), True)

# forms

_afs = uicfg.autoform_section
_affk = uicfg.autoform_field_kwargs
_afs.tag_subject_of(("CWUser", "lives_at", "*"), "main", "inlined")
_afs.tag_subject_of(("Expense", "has_lines", "*"), "main", "inlined")

_afs.tag_subject_of(("ExpenseLine", "paid_by", "*"), "main", "attributes")
_afs.tag_subject_of(("ExpenseLine", "paid_by", "*"), "muledit", "attributes")
_affk.tag_subject_of(("ExpenseLine", "paid_by", "*"), {"sort": True})
_afs.tag_subject_of(("ExpenseLine", "paid_for", "*"), "main", "attributes")
_affk.tag_subject_of(("ExpenseLine", "paid_for", "*"), {"sort": True})

# XXX still necessary?
uicfg.autoform_permissions_overrides.tag_subject_of(
    ("Expense", "has_lines", "*"), "add_on_new"
)


class RefundChangeStateForm(workflow.ChangeStateForm):
    __select__ = is_instance("Refund")
    payment_date = autoform.etype_relation_field("Refund", "payment_date")
    payment_mode = autoform.etype_relation_field("Refund", "payment_mode")


# views

_pvs = uicfg.primaryview_section
_pvs.tag_attribute(("Expense", "title"), "hidden")
_pvs.tag_subject_of(("Expense", "has_lines", "*"), "hidden")
_pvs.tag_subject_of(("Expense", "has_attachment", "File"), "hidden")
_pvs.tag_subject_of(("Refund", "has_lines", "*"), "hidden")
_pvdc = uicfg.primaryview_display_ctrl
_pvdc.tag_attribute(("Expense", "description"), {"showlabel": False})


class ExpensePrimaryView(primary.PrimaryView):
    __select__ = is_instance("Expense")

    def render_entity_title(self, entity):
        state = self._cw._(entity.cw_adapt_to("IWorkflowable").state)
        title = xml_escape("%s - %s" % (entity.dc_title(), state))
        self.w(
            '<h1><span class="etype">%s</span> %s</h1>'
            % (entity.dc_type().capitalize(), title)
        )


class ExpenseSummaryComponent(component.EntityCtxComponent):
    __regid__ = "expense.summary"
    __select__ = component.EntityCtxComponent.__select__ & is_instance("Expense")
    context = "navcontentbottom"
    title = _("Expense summary")
    order = -1

    def field(self, label, value, w):
        w('<div class="entityfield">')
        w('<span class="label">%s</span> ' % label)
        w("<span>%s</span></div>" % value)

    def render_body(self, w):
        _ = self._cw._
        entity = self.entity
        total = "%s %s %s" % (
            entity.euro_total(),
            _("including taxes"),
            entity.euro_taxes(),
        )
        self.field(_("total"), total, w)
        rset = self._cw.execute(
            "Any EID,T,ET,EA,EC,C,GROUP_CONCAT(CCL),CL "
            "GROUPBY EID,T,ET,EC,EA,C,CL "
            "WHERE X has_lines E, X eid %(x)s, E eid EID, "
            "E type T, E title ET, E currency EC, "
            "E amount EA, E paid_by C?, C label CL, "
            "E paid_for CC, CC label CCL",
            {"x": entity.eid},
        )
        self._cw.view("expense.table", rset, w=w)


class ExpenseTable(tableview.RsetTableView):
    __regid__ = "expense.table"
    headers = [
        _("eid"),
        _("type"),
        _("title"),
        _("amount"),
        _("currency"),
        _("paid_by"),
        _("paid_for"),
    ]
    layout_args = {"display_filter": "top"}


class RefundPrimaryView(primary.PrimaryView):
    __select__ = is_instance("Refund")

    def render_entity_title(self, entity):
        state = self._cw._(entity.cw_adapt_to("IWorkflowable").state)
        title = xml_escape("%s - %s" % (entity.dc_title(), state))
        self.w(
            '<h1><span class="etype">%s</span> %s</h1>'
            % (entity.dc_type().capitalize(), title)
        )


class RefundSummaryComponent(ExpenseSummaryComponent):
    __regid__ = "expense.summary"
    __select__ = component.EntityCtxComponent.__select__ & is_instance("Refund")
    title = _("Refund summary")

    def render_body(self, w):
        _ = self._cw._
        entity = self.entity
        accounts = [act.view("oneline") for act in entity.paid_by_accounts()]
        self.field(_("account to refund"), ",".join(accounts), w)
        self.field(_("total"), entity.total, w)
        rset = self._cw.execute(
            "Any E,ET,EC,EA WHERE X has_lines E, X eid %(x)s, "
            "E title ET, E currency EC, E amount EA",
            {"x": entity.eid},
        )
        self._cw.view("table", rset, w=w)


try:
    from cubicweb_expense.pdfgen.writers import PDFWriter

    has_reportlab = yes()
except ImportError:
    has_reportlab = yes(0)

# use the has_reportlab selector trick to have them in the registry anyway
# and avoid for instance i18n messages removing due to missing reportlab


class PDFAction(action.Action):
    __regid__ = "pdfaction"
    __select__ = has_reportlab & one_line_rset() & is_instance("Expense", "Refund")

    title = _("generate pdf document")
    category = "mainactions"

    def url(self):
        return self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0).absolute_url(
            vid="pdfexport"
        )


class PDFExportView(EntityView):
    __regid__ = "pdfexport"
    __select__ = has_reportlab & one_line_rset() & is_instance("Refund", "Expense")

    title = _("pdf export")
    content_type = "application/pdf"
    templatable = False
    binary = True

    def cell_call(self, row, col):
        # import error to avoid import error if reportlab isn't available
        _ = self._cw._
        writer = PDFWriter(self._cw.vreg.config)
        entity = self.cw_rset.get_entity(row, col)
        entity.complete()
        # XXX reportlab needs HOME and getcwd to find fonts
        home_backup = os.environ.get("HOME")
        getcwd_backup = os.getcwd
        try:
            os.environ["HOME"] = "wtf"
            os.getcwd = lambda: "wtf"
            # NOTE: we could use self.w.__self__ directly
            stream = BytesIO()
            writer.write(entity, stream)
            self.w(stream.getvalue())
        finally:
            if home_backup:
                os.environ["HOME"] = home_backup
            os.getcwd = getcwd_backup


# boxes ########################################################################


class ExpenseLineFilesComponent(component.EntityCtxComponent):
    """display the list of files attached to the expenselines"""

    __regid__ = "expenseline.attachments"
    __select__ = (
        component.EntityCtxComponent.__select__
        & authenticated_user()
        & is_instance("Expense")
        & rql_condition(
            "EXISTS(X has_lines Y, Y has_attachment F) " "OR EXISTS(X has_attachment F)"
        )
    )
    title = _("has_attachment")

    def render_body(self, w):
        rset = self._cw.execute(
            "(Any F WHERE X has_attachment F, X eid %(x)s)"
            " UNION "
            "(Any F WHERE X has_lines Y, Y has_attachment F, X eid %(x)s)",
            dict(x=self.entity.eid),
        )
        self._cw.view("incontext", rset, "null", w=w)



# -*- coding: iso-8859-15 -*-

"""
Library package containing the definition of the class that can write a PDF
file for Fresh documents (Refund or Expense). The PDF generation uses
ReportLab platypus and is based on specific flow generators and
templates.
"""

from cubicweb_expense.pdfgen.templates import FreshDocTemplate
from cubicweb_expense.pdfgen.flowgenerators import ExpenseFlowGenerator
from cubicweb_expense.pdfgen.flowgenerators import RefundFlowGenerator


class PDFWriter(object):
    """
    Class writing the PDF for the Fresh documents (Refund or Expense). This
    class instanciates a flow generator specific to the kind of document
    (Expense or Refund) that must be written.
    """

    def __init__(self, company_data):
        """
        Instanciates a PDF writer.

        company_data: {'':u"", }. dictionnary containing various data about
                      the company.
        """
        self.company_data = company_data

        # flow generator that can create the PDF content flow of the document.
        self._flow_generator = None  # AbstractFlowGenerator
        # type of document to be generated
        self._doc_type = None  # string "expense" or "refund"

    def __init_flow_generator(self, entity):
        """
        Instanciates the flow generator that will create the PDF flow content
        of the document. The real class of flow generator
        (ExpenseFlowGenerator or RefundFlowGenerator) depends on the class
        of the entity object.

        entity: Entity. Any object of class Refund or Expense.
        """
        if entity.e_schema == "Refund":
            if self._doc_type != "refund":
                self._doc_type = "refund"
                self._flow_generator = RefundFlowGenerator()
        elif entity.e_schema == "Expense":
            if self._doc_type != "expense":
                self._doc_type = "expense"
                self._flow_generator = ExpenseFlowGenerator()
        else:
            self._doc_type = None
            self._flow_generator = None

    def write(self, entity, output):
        """
        Writes a PDF document representing the entity object.

        entity: Entity. Any object of class Refund or Expense.
        output: string or output flow. File name or IO flow that the PDF will
                be written in.
        """
        # Initializes a flow generator (whose class depends on the entity type)
        self.__init_flow_generator(entity)

        if self._doc_type is None:
            err_msg = (
                "Entity #%d can not be displayed as a PDF file. "
                + 'Should be of type "Refund" or "Expense".' % entity.eid
            )
            raise Exception(err_msg)

        # Opens the PDF document
        pdf_doc = FreshDocTemplate(
            output, self._doc_type, self.company_data, entity._cw._
        )

        # Builds it with the content flow returned by the flow generator
        pdf_doc.build(self._flow_generator.generate_flow(entity))



"""this contains the template-specific entities' classes"""

from cubicweb.entities import AnyEntity, fetch_config


class LineContainerMixIn(object):
    """mixin class used by all entities containing expense lines"""

    def paid_by(self):
        """returns the list of eusers who paid something
        in this container
        """
        accounts = self.paid_by_accounts()
        return [acc.associated_to[0] for acc in accounts if acc.associated_to]

    def paid_by_accounts(self):
        """returns the list of accounts used in this container"""
        accounts = []
        eids = set()
        for line in self.has_lines:
            account = line.paid_by[0]
            if account.eid not in eids:
                accounts.append(account)
                eids.add(account.eid)
        return accounts

    def totals_paid_by(self):
        """returns a dictionnary containing the total paid by each euser who
        paid something
        """
        tot = {}
        for line in self.has_lines:
            account = line.paid_by[0]
            if account.associated_to:
                euser = account.associated_to[0]
                tot.setdefault(euser, 0.0)
                tot[euser] += line.euro_amount()
        return tot

    @property
    def start(self):
        return min(line.diem for line in self.has_lines)

    @property
    def stop(self):
        return max(line.diem for line in self.has_lines)

    @property
    def total(self):
        return sum(line.amount for line in self.has_lines)

    @property
    def taxes(self):
        return sum(line.taxes for line in self.has_lines)


class Expense(LineContainerMixIn, AnyEntity):
    __regid__ = "Expense"
    fetch_attrs, cw_fetch_order = fetch_config(["title"])

    def dc_long_title(self):
        users = self.paid_by()
        if users:
            return "%s (%s)" % (
                self.title,
                ", ".join(euser.login for euser in self.paid_by()),
            )
        return self.title

    def euro_taxes(self):
        return sum(line.euro_taxes() for line in self.has_lines)

    def euro_total(self):
        return sum(line.euro_amount() for line in self.has_lines)


class ExpenseLine(AnyEntity):
    __regid__ = "ExpenseLine"
    fetch_attrs, cw_fetch_order = fetch_config(
        ["diem", "type", "title", "amount", "currency"]
    )

    @property
    def parent_expense(self):
        expenses = [
            entity for entity in self.reverse_has_lines if entity.e_schema == "Expense"
        ]
        if expenses:
            return expenses[0]
        return None

    def dc_title(self):
        return "%s - %s - %s - %s %s" % (
            self._cw.format_date(self.diem),
            self._cw._(self.type),
            self.title,
            self.amount,
            self.currency,
        )

    def dc_long_title(self):
        expense = self.parent_expense
        if expense:
            return "%s - %s" % (self.title, expense.dc_title())
        else:
            return self.dc_title()

    def euro_amount(self):
        if self.currency == "EUR":
            return self.amount
        return self.exchange_rate * self.amount

    def euro_taxes(self):
        if self.taxes_currency == "EUR":
            return self.taxes
        return self.taxes_exchange_rate * self.taxes


class PaidByAccount(AnyEntity):
    __regid__ = "PaidByAccount"
    fetch_attrs, cw_fetch_order = fetch_config(["label", "account"])

    def dc_title(self):
        return self.label

    def dc_long_title(self):
        return "%s (%s)" % (self.label, self.account)


class PaidForAccount(PaidByAccount):
    __regid__ = "PaidForAccount"


class Refund(LineContainerMixIn, AnyEntity):
    __regid__ = "Refund"
    fetch_attrs = ("payment_date", "payment_mode")

    def dc_title(self):
        _ = self._cw._
        return "%s %s, %s: %s" % (
            _("refund for account"),
            self.paid_by_accounts()[0].view("textincontext"),
            _("amount"),
            self.total,
        )

    def for_user(self):
        account = self.to_account[0]
        if account.associated_to:
            return account.associated_to[0]
        return None



import unittest

from _helpers import HelpersTC


class EntitiesTC(HelpersTC):
    def test_totals_paid_by(self):
        with self.admin_access.repo_cnx() as cnx:
            expense = cnx.create_entity("Expense", title="expense 2")
            self.add_expense_line(cnx, expense, self.account1)
            self.add_expense_line(cnx, expense, self.account2)
            self.add_expense_line(cnx, expense, self.account1)
            paid_by = dict(
                (euser.eid, value) for euser, value in expense.totals_paid_by().items()
            )
            self.assertEqual(paid_by, {self.user1: 2, self.user2: 1})


if __name__ == "__main__":
    unittest.main()



"""unittests for cubicweb-expense hooks"""

import unittest

from cubicweb.devtools.testlib import MAILBOX

from _helpers import HelpersTC


class HooksTC(HelpersTC):
    def test_refund_is_created(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any R,PA,COUNT(L) WHERE R has_lines L, R to_account PA")
            # only one refund should have been created (because no refund
            # is needed for company expenses)
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][1], self.account1)
            self.assertEqual(rset[0][2], 1)

    def test_refunds_gets_updated_on_new_lines(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any R,PA WHERE R has_lines L, R to_account PA")
            refund = rset[0][0]
            expense = cnx.create_entity("Expense", title="expense 2")
            self.add_expense_line(cnx, expense, self.account1)
            cnx.commit()
            self.accept(cnx, expense)
            rset = cnx.execute(
                "Any COUNT(L) WHERE R has_lines L, R eid %(r)s", {"r": refund}
            )
            self.assertEqual(rset[0][0], 2)

    def test_no_refund_is_created_for_company_account(self):
        with self.admin_access.repo_cnx() as cnx:
            count = cnx.execute(
                'Any COUNT(R) WHERE R is Refund, R in_state S, S name "preparation"'
            )[0][0]
            expense = cnx.create_entity("Expense", title="company expense")
            self.add_expense_line(cnx, expense, self.account_comp)
            cnx.commit()
            self.accept(cnx, expense)
            newcount = cnx.execute(
                'Any COUNT(R) WHERE R is Refund, R in_state S, S name "preparation"'
            )[0][0]
            self.assertEqual(newcount, count)

    def test_no_refund_is_created_while_not_accepted(self):
        """make sure no refund is created until expense was accepted"""
        with self.admin_access.repo_cnx() as cnx:
            count = self.refund_lines_count(cnx, self.account1)
            expense = cnx.create_entity("Expense", title="expense 1")
            self.add_expense_line(cnx, expense, self.account1)
            cnx.commit()  # to fire corresponding operations
            newcount = self.refund_lines_count(cnx, self.account1)
            self.assertEqual(newcount, count)
            self.accept(cnx, expense)
            newcount = self.refund_lines_count(cnx, self.account1)
            self.assertEqual(newcount, count + 1)

    def test_expense_accepted_notification(self):
        with self.admin_access.repo_cnx() as cnx:
            expense = cnx.create_entity("Expense", title="expense 2")
            self.add_expense_line(cnx, expense, self.account1)
            # force expense to its initial state, otherwise StatusChangeHook won't be called
            cnx.commit()
            self.assertEqual(len(MAILBOX), 0, MAILBOX)
            self.accept(cnx, expense)  # to fire corresponding operations
            self.assertEqual(len(MAILBOX), 1, MAILBOX)
            self.assertCountEqual(MAILBOX[0].recipients, ["john@test.org"])

    def test_refund_acted_notification(self):
        with self.admin_access.repo_cnx() as cnx:
            expense1 = cnx.create_entity("Expense", title="expense 2")
            self.add_expense_line(cnx, expense1, self.account1)
            cnx.commit()
            self.accept(cnx, expense1)
            MAILBOX[:] = []
            account1 = cnx.execute(
                "Any X WHERE X eid %(x)s", {"x": self.account1}
            ).get_entity(0, 0)
            account1.reverse_to_account[0].cw_adapt_to("IWorkflowable").fire_transition(
                "pay"
            )
            self.assertEqual(len(MAILBOX), 0, MAILBOX)
            cnx.commit()  # to fire corresponding operations
            self.assertEqual(len(MAILBOX), 1, MAILBOX)
            email1 = MAILBOX[0]
            self.assertCountEqual(email1.recipients, ["john@test.org"])
            MAILBOX[:] = []
            expense2 = cnx.create_entity("Expense", title="expense 3")
            self.add_expense_line(cnx, expense2, self.account1)
            cnx.commit()
            self.accept(cnx, expense2)
            account1.cw_clear_all_caches()
            account1.reverse_to_account[0].cw_adapt_to("IWorkflowable").fire_transition(
                "pay"
            )
            cnx.commit()  # to fire corresponding operations
            email2 = MAILBOX[0]
            self.assertCountEqual(email2.recipients, ["john@test.org"])
            self.assertNotEqual(
                email2.message.get("Message-id"), email1.message.get("Message-id")
            )

    def test_automatic_refund_with_existing_line(self):
        with self.admin_access.repo_cnx() as cnx:
            refund = cnx.create_entity("Refund")
            # NOTE: use account2 which doesn't have a refund yet
            self.add_relation(cnx, refund.eid, "to_account", self.account2)
            expense = cnx.create_entity("Expense", title="expense 2")
            line1eid = self.add_expense_line(cnx, expense, self.account2)
            line2eid = self.add_expense_line(cnx, expense, self.account2)
            self.add_expense_line(cnx, expense, self.account2)
            self.add_relation(cnx, refund.eid, "has_lines", line1eid)
            self.add_relation(cnx, refund.eid, "has_lines", line2eid)
            cnx.commit()
            rset = cnx.execute(
                "Any R,COUNT(L) WHERE R has_lines L, R eid %s" % refund.eid
            )
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][1], 2)
            self.accept(cnx, expense)
            rset = cnx.execute(
                "Any L WHERE R has_lines L, R eid %(r)s", {"r": refund.eid}
            )
            self.assertEqual(len(rset), 3)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
try:
    from cubicweb import server

    server.ON_COMMIT_ADD_RELATIONS.add("has_lines")
except ImportError:
    pass

options = (
    (
        "logo-filename",
        {
            "type": "string",
            "default": "/some/path/to/logo.png",
            "help": "path to the logo used to build pdfs",
            "group": "expense",
            "level": 2,
        },
    ),
    (
        "company-name",
        {
            "type": "string",
            "default": "MyCompany",
            "help": "name of the company editing PDFs",
            "group": "expense",
            "level": 2,
        },
    ),
    (
        "company-address",
        {
            "type": "string",
            "default": "1, Rue du chat qui pêche - 75005 Paris",
            "help": "address of the company editing PDFs",
            "group": "expense",
            "level": 2,
        },
    ),
    (
        "company-offnum",
        {
            "type": "string",
            "default": "123.456.789.00011",
            "help": "official ID number of the company editing PDFs",
            "group": "expense",
            "level": 2,
        },
    ),
    (
        "company-actnum",
        {
            "type": "string",
            "default": "101AZ",
            "help": "activity number of the company editing PDFs",
            "group": "expense",
            "level": 2,
        },
    ),
)



"""tests for expense's security"""

import unittest

from cubicweb import Unauthorized, ValidationError

from _helpers import HelpersTC


class SecurityTC(HelpersTC):
    def test_users_cannot_modify_accepted_expense(self):
        with self.new_access("john").repo_cnx() as cnx:
            rset = cnx.execute(
                "Any E WHERE E is Expense, E has_lines EE, EE paid_by PA, "
                'E in_state S, S name "accepted", PA associated_to U, U eid %(u)s',
                {"u": self.user1},
            )  # user1 is john
            self.assertEqual(len(rset), 1)
            expense = rset.get_entity(0, 0)
            self.add_expense_line(cnx, expense, self.account1)
            self.assertRaises(Unauthorized, cnx.commit)

    def test_users_cannot_accept_expense(self):
        with self.new_access("john").repo_cnx() as cnx:
            expense = self.create_and_submit_expense(cnx)
            cnx.commit()
            self.assertRaises(
                ValidationError,
                expense.cw_adapt_to("IWorkflowable").fire_transition,
                "accept",
            )

    def test_users_cannot_update_accepted_expense_line(self):
        with self.admin_access.repo_cnx() as cnx:
            expense = cnx.create_entity("Expense", title="company expense")
            lineeid = self.add_expense_line(cnx, expense, self.account1)
            cnx.commit()
            self.accept(cnx, expense)
        with self.new_access("john").repo_cnx() as cnx:
            cnx.execute(
                "SET E amount %(a)s WHERE E eid %(e)s", {"e": lineeid, "a": 12.3}
            )
            self.assertRaises(Unauthorized, cnx.commit)

    def test_users_can_create_expenses(self):
        with self.new_access("john").repo_cnx() as cnx:
            self.create_and_submit_expense(cnx)
            cnx.commit()

    def test_users_cannot_create_refunds(self):
        with self.new_access("john").repo_cnx() as cnx:
            rset = cnx.execute(
                "Any EE WHERE E is Expense, E has_lines EE, EE paid_by PA, "
                'E in_state S, S name "accepted", PA associated_to U, U eid %(u)s',
                {"u": self.user1},
            )  # user1 is john
            self.assertEqual(len(rset), 1)
            lineeid = rset[0][0]
            rql = "INSERT Refund R: R has_lines E, R to_account A WHERE E eid %(e)s, A eid %(a)s"
            cnx.execute(rql, {"e": lineeid, "a": self.account1})
            self.assertRaises(Unauthorized, cnx.commit)

    def test_users_cannot_update_refunds(self):
        with self.new_access("john").repo_cnx() as cnx:
            rset = cnx.execute(
                "Any R WHERE R is Refund, R has_lines EE, EE paid_by PA, "
                "R to_account PA, PA associated_to U, U eid %(u)s",
                {"u": self.user1},
            )  # user1 is john
            self.assertEqual(len(rset), 1)
            line = self.new_expense_line(cnx, self.account1)
            rql = "SET R has_lines E WHERE E eid %(e)s"
            cnx.execute(rql, {"e": line.eid})
            self.assertRaises(Unauthorized, cnx.commit)


if __name__ == "__main__":
    unittest.main()



# -*- coding: iso-8859-15 -*-

"""
Library package containing the definition of flow generation classes used in
the PDF generation with ReportLab platypus. These classes generate the flow
content for an entity Refund or Expense in the Fresh application. This
flow content will be displayed on various pages by the FreshDocTemplate
through the platypus machinery.
"""

from reportlab.lib.units import cm
from reportlab.lib.colors import black

from reportlab.platypus import Paragraph, Spacer, KeepTogether
from reportlab.platypus import Table, TableStyle

from logilab.mtconverter import xml_escape

from cubicweb_expense.pdfgen.utils import format_number
from cubicweb_expense.pdfgen.styles import style_metadata, style_ids, style_address
from cubicweb_expense.pdfgen.styles import style_table_header, style_table_data
from cubicweb_expense.pdfgen.styles import style_table_data_num


class AbstractFlowGenerator:
    """
    Abstract class containing the common parts for generating the PDF flow
    content for the various kinds of documents generated by the Fresh
    application (Refund, Expense). Specific parts are generated in the
    derived classes (one for each kind of document) (cf. Template design
    pattern).
    For a correct behaviour, you must use a flow generator corresponding to
    the kind of document you want to render in PDF.
    The PDF flow content is a list of Flowable that has to be displayed on
    various pages.
    """

    def __init__(self):
        """
        Instanciates the flow generator. This constructor should never be
        called directly as this class is abstract, it will be called by a
        child class constructor.
        """
        if self.__class__ is AbstractFlowGenerator:
            raise NotImplementedError()

    def generate_flow(self, entity):
        """
        Generates the flow content (a list of Flowable) for the entity.

        entity: Entity. Any entity of class Refund or Expense that must
                be displayed in PDF.
        Returns: [Flowable, ...]
        """
        flow_list = []  # [Flowable, ...]

        # top table containing on the left the metadata and on the right the
        # name of the person concerned by the document
        top_table = Table(
            [
                [self.get_metadata(entity), "", self.get_person_name(entity)],
            ],
            [11.5 * cm, 0.5 * cm, 7 * cm],
        )
        top_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0 * cm),
                ]
            )
        )

        flow_list.append(top_table)

        flow_list.append(Spacer(1, 0.5 * cm))

        # main table containing the expenses description
        flow_list.append(self.get_expenses_table(entity))

        flow_list.append(Spacer(1, 1 * cm))

        # bottom table containing on the right the total table and on the left
        # other useful data (payment, subtotals)
        bottom_table = Table(
            [
                [self.get_total_relative_data(entity), " ", self.get_total(entity)],
            ],
            [12.5 * cm, 0.5 * cm, 6 * cm],
        )
        bottom_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0 * cm),
                ]
            )
        )

        flow_list.append(KeepTogether(bottom_table))

        return flow_list

    def get_metadata(self, entity):
        """
        Gives the table that contains metadata about the entity (title,
        workcase, etc.)
        This abstract method will be defined in the children classes.

        entity: Entity. Any entity of class Expense or Refund.
        Returns: Flowable (very likely a Table).
        """
        raise NotImplementedError()

    def get_person_name(self, entity):
        """
        Gives the table that contains the name and address of the person
        concerned by th document.
        This abstract method will be defined in the children classes.

        entity: Entity. Any entity of class Expense or Refund.
        Returns: Flowable (very likely a Table).
        """
        raise NotImplementedError()

    def get_total(self, entity):
        """
        Gives the table that contains the total.
        This abstract method will be defined in the children classes.

        entity: Entity. Any entity of class Expense or Refund.
        Returns: Flowable (very likely a Table).
        """
        raise NotImplementedError()

    def get_total_relative_data(self, entity):
        """
        Gives the table that contains useful data relative to the total
        (subtotals, payment information, etc.)
        This abstract method will be defined in the children classes.

        entity: Entity. Any entity of class Expense or Refund.
        Returns: Flowable (very likely a Table).
        """
        raise NotImplementedError()

    def get_expenses_table(self, entity):
        """
        Gives the main table of the document (that contains lines describing
        the actual expenses).
        This abstract method will be defined in the children classes.

        entity: Entity. Any entity of class Expense or Refund.
        Returns: Flowable (very likely a Table).
        """
        raise NotImplementedError()


class ExpenseFlowGenerator(AbstractFlowGenerator):
    """
    Class building the specific contents for PDF flow content generation.
    This class must be used when we want to generate the flow for an
    Expense.
    """

    def __init__(self):
        """
        Initializes the flow content generator for Expense entities.
        """
        AbstractFlowGenerator.__init__(self)

    def get_metadata(self, entity):
        """
        Gives the table that contains metadata about the entity (title,
        workcase, etc.)

        entity: Entity. Any entity of class Expense.         .
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        col_data = []

        s = "%s%s%s" % (
            xml_escape(_("Subject")),
            xml_escape(_(": ")),
            xml_escape(entity.title),
        )
        col_data.append(Paragraph(s, style_metadata))

        try:
            wcase = entity.spent_for[0]
        except (IndexError, AttributeError, KeyError):
            pass
        else:
            s = xml_escape(wcase.ref)
            s = "%s%s%s" % (
                xml_escape(_("Workcase")),
                xml_escape(_(": ")),
                xml_escape(wcase.ref),
            )
            col_data.append(Paragraph(s, style_metadata))

        s = "%s%s#%s" % (xml_escape(_("Reference")), xml_escape(_(": ")), entity.eid)
        col_data.append(Paragraph(s, style_metadata))

        meta_table = Table(
            [
                [col_data],
            ],
            [11.5 * cm],
        )
        meta_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                ]
            )
        )

        return meta_table

    def get_person_name(self, entity):
        """
        Gives the table that contains the name and address of the person
        concerned by th document.

        entity: Entity. Any entity of class Expense.
        Returns: Flowable (very likely a Table).
        """
        return Spacer(0, 0 * cm)

    def get_total(self, entity):
        """
        Gives the table that contains the total.

        entity: Entity. Any entity of class Expense.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        tab_data = []

        # First line
        lin_data = []
        s = "<b>%s</b>" % xml_escape(_("Total (EUR)"))
        lin_data.append(Paragraph(s, style_table_data))

        s = "<b>%s</b>" % format_number(entity.total, 2)
        lin_data.append(Paragraph(s, style_table_data_num))

        tab_data.append(lin_data)

        # Second line
        lin_data = []
        s = "<i>%s</i>" % xml_escape(_("included taxes (EUR)"))
        lin_data.append(Paragraph(s, style_table_data))

        s = "<i>%s</i>" % format_number(entity.euro_taxes(), 2)
        lin_data.append(Paragraph(s, style_table_data_num))

        tab_data.append(lin_data)

        tot_table = Table(tab_data, [3.25 * cm, 2.75 * cm])
        tot_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                    ("INNERGRID", (0, 0), (-1, -1), 0.02 * cm, black),
                ]
            )
        )

        return tot_table

    def get_total_relative_data(self, entity):
        """
        Gives the table that contains useful data relative to the total
        (subtotals, payment information, etc.)

        entity: Entity. Any entity of class Expense.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        col_data = []

        refundable_tot = entity.totals_paid_by()
        for euser in refundable_tot.keys():
            s = "%s %s %s %s" % (
                format_number(refundable_tot[euser], 2),
                xml_escape(_("EUR")),
                xml_escape(_("to be refunded to")),
                xml_escape(euser.dc_title()),
            )
            col_data.append(Paragraph(s, style_table_data))

        if col_data != []:
            rel_table = Table(
                [
                    [col_data],
                ],
                [12.5 * cm],
            )
            rel_table.setStyle(
                TableStyle(
                    [
                        ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                        ("VALIGN", (0, 0), (-1, -1), "TOP"),
                        ("LEFTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                        ("RIGHTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                        ("TOPPADDING", (0, 0), (-1, -1), 0.2 * cm),
                        ("BOTTOMPADDING", (0, 0), (-1, -1), 0.2 * cm),
                        ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                    ]
                )
            )

            return rel_table

        else:
            return Spacer(0, 0 * cm)

    def get_expenses_table(self, entity):
        """
        Gives the main table of the document (that contains lines describing
        the actual expenses).

        This table contains only one column. In this column, a table is
        inserted for each expense line. This table also contains one column.
        In this column, we insert two tables (one per line) because each line
        has different column sizes. See the _build_inner_tables method for
        details.

        entity: Entity. Any entity of class Expense.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        tab_data = []

        # Writes the header lines

        # First line
        lin_data_1 = []

        s = "<b>%s</b>" % xml_escape(_("Date"))
        lin_data_1.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Description"))
        lin_data_1.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Expense Type"))
        lin_data_1.append(Paragraph(s, style_table_header))

        # Second line
        lin_data_2 = []

        s = "<b>%s</b>" % xml_escape(_("Paid by"))
        lin_data_2.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Paid for"))
        lin_data_2.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Amount"))
        lin_data_2.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Curr."))
        lin_data_2.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Rate"))
        lin_data_2.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Amount (EUR)"))
        lin_data_2.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Taxes (EUR)"))
        lin_data_2.append(Paragraph(s, style_table_header))

        # adds a table built with the two previously defined lines
        tab_data.append([self._build_inner_tables(lin_data_1, lin_data_2)])

        # Writes the accounting lines
        for exp_line in entity.has_lines:
            # First line
            lin_data_1 = []

            s = xml_escape(exp_line.printable_value("diem"))
            lin_data_1.append(Paragraph(s, style_table_data))

            s = xml_escape(exp_line.title)
            lin_data_1.append(Paragraph(s, style_table_data))

            s = xml_escape(_(exp_line.type))
            lin_data_1.append(Paragraph(s, style_table_data))

            # Second line
            lin_data_2 = []

            s = xml_escape(exp_line.paid_by[0].label)
            lin_data_2.append(Paragraph(s, style_table_data))

            col_data = []
            for dest in exp_line.paid_for:
                s = xml_escape(dest.label)
                col_data.append(Paragraph(s, style_table_data))
            lin_data_2.append(col_data)

            s = format_number(exp_line.amount, 2)
            lin_data_2.append(Paragraph(s, style_table_data_num))

            s = xml_escape(exp_line.currency)
            lin_data_2.append(Paragraph(s, style_table_data))

            s = format_number(exp_line.exchange_rate, 5)
            lin_data_2.append(Paragraph(s, style_table_data_num))

            s = "<b>%s</b>" % format_number(exp_line.euro_amount(), 2)
            lin_data_2.append(Paragraph(s, style_table_data_num))

            s = "<i>%s</i>" % format_number(exp_line.euro_taxes(), 2)
            lin_data_2.append(Paragraph(s, style_table_data_num))

            # adds a table built with the two previously defined lines
            tab_data.append([self._build_inner_tables(lin_data_1, lin_data_2)])

        # Builds the main table that will be returned
        main_table = Table(tab_data, [19 * cm])
        main_table.repeatRows = 1
        main_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -2), 0.3 * cm),
                    ("BOTTOMPADDING", (0, -1), (-1, -1), 0 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                ]
            )
        )

        return main_table

    def _build_inner_tables(self, line1_data, line2_data):
        """
        Builds inner tables that will be displayed into the main table (the
        one that contains ExpenseLines). Hence, an ExpenseLine cannot be
        displayed on a single line, thus we build two tables of one line to
        display one ExpenseLine. Each table has a different number of columns
        and different sizes of columns.

        line1_data: [Flowable, ...] first line
        line2_data: [Flowable, ...] second line
        """
        # Builds table for the first line
        line1_table = Table([line1_data], [2.25 * cm, 14 * cm, 2.75 * cm])
        line1_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, 0), "LEFT"),
                    ("VALIGN", (0, 0), (-1, 0), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("TOPPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("INNERGRID", (0, 0), (-1, 0), 0.02 * cm, black),
                ]
            )
        )

        # Builds table for the second line
        line2_table = Table(
            [line2_data],
            [
                4.25 * cm,
                4.5 * cm,
                2.25 * cm,
                1.25 * cm,
                1.75 * cm,
                2.75 * cm,
                2.25 * cm,
            ],
        )
        line2_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, 0), "LEFT"),
                    ("VALIGN", (0, 0), (-1, 0), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("TOPPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0.1 * cm),
                    ("INNERGRID", (0, 0), (-1, 0), 0.02 * cm, black),
                    ("LINEABOVE", (0, 0), (-1, 0), 0.02 * cm, black),
                ]
            )
        )

        # Builds the table that contains the two lines describing the expense
        both_lines_table = Table([[line1_table], [line2_table]], [19 * cm])
        both_lines_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.02 * cm, black),
                ]
            )
        )

        return both_lines_table


class RefundFlowGenerator(AbstractFlowGenerator):
    """
    Class building the specific contents for PDF flow content generation.
    This class must be used when we want to generate the flow for a
    Refund.
    """

    def __init__(self):
        """
        Initializes the flow content generator for Refund entities.
        """
        AbstractFlowGenerator.__init__(self)

    def get_metadata(self, entity):
        """
        Gives the table that contains metadata about the entity (title,
        workcase, etc.)

        entity: Entity. Any entity of class Refund.
        Returns: Flowable (very likely a Table).
        """
        return Spacer(0, 0 * cm)

    def get_person_name(self, entity):
        """
        Gives the table that contains the name and address of the person
        concerned by th document.

        entity: Entity. Any entity of class Refund.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        col_data = []

        euser = entity.for_user()
        if euser:
            s = "%s%s%s" % (
                xml_escape(_("SS number")),
                xml_escape(_(": ")),
                xml_escape(euser.ssnum or "missing SS number"),
            )
            col_data.append(Paragraph(s, style_ids))

            col_data.append(Spacer(1, 0.2 * cm))

            s = "<b>%s</b>" % xml_escape(euser.name())
            col_data.append(Paragraph(s, style_address))

            if euser.lives_at:
                addr = euser.lives_at[0]

                s = xml_escape(addr.street)
                col_data.append(Paragraph(s, style_address))

                s = "%s %s" % (xml_escape(addr.postalcode), xml_escape(addr.city))
                col_data.append(Paragraph(s, style_address))

        else:  # could not find related euser
            s = "<b>???</b>"
            col_data.append(Paragraph(s, style_address))

        # Builds address table
        addr_table = Table(
            [
                [col_data],
            ],
            [7 * cm],
        )
        addr_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                ]
            )
        )

        return addr_table

    def get_total(self, entity):
        """
        Gives the table that contains the total.

        entity: Entity. Any entity of class Refund.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        lin_data = []

        s = "<b>%s</b>" % xml_escape(_("To refund (EUR)"))
        lin_data.append(Paragraph(s, style_table_data))

        s = "<b>%s</b>" % format_number(entity.total, 2)
        lin_data.append(Paragraph(s, style_table_data_num))

        # Builds the total table
        tot_table = Table([lin_data], [3.25 * cm, 2.75 * cm])
        tot_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                    ("INNERGRID", (0, 0), (-1, -1), 0.02 * cm, black),
                ]
            )
        )

        return tot_table

    def get_total_relative_data(self, entity):
        """
        Gives the table that contains useful data relative to the total
        (subtotals, payment information, etc.)

        entity: Entity. Any entity of class Refund.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        col_data = []

        if entity.payment_date is None:
            date = "\u00A0 \u00A0 \u00A0 \u00A0 \u00A0 \u00A0 \u00A0 "
        else:
            date = entity.printable_value("payment_date")
        if entity.payment_mode is None:
            mode = ""
        else:
            mode = entity.payment_mode
        s = "%s %s %s %s" % (
            xml_escape(_("Refunded on")),
            xml_escape(date),
            xml_escape(_("by")),
            xml_escape(mode),
        )
        col_data.append(Paragraph(s, style_table_data))

        # Builds payment table
        payment_table = Table(
            [
                [col_data],
            ],
            [12.5 * cm],
        )
        payment_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0.2 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                ]
            )
        )

        return payment_table

    def get_expenses_table(self, entity):
        """
        Gives the main table of the document (that contains lines describing
        the actual expenses).

        entity: Entity. Any entity of class Refund.
        Returns: Flowable (very likely a Table).
        """
        _ = entity._cw._
        tab_data = []

        # Writes the header line
        lin_data = []

        s = "<b>%s</b>" % xml_escape(_("Date"))
        lin_data.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Description"))
        lin_data.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Workcase"))
        lin_data.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Amount"))
        lin_data.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Curr."))
        lin_data.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Rate"))
        lin_data.append(Paragraph(s, style_table_header))

        s = "<b>%s</b>" % xml_escape(_("Amount (EUR)"))
        lin_data.append(Paragraph(s, style_table_header))

        tab_data.append(lin_data)

        # Writes the accounting lines
        for exp_line in entity.has_lines:
            lin_data = []

            s = xml_escape(exp_line.printable_value("diem"))
            lin_data.append(Paragraph(s, style_table_data))

            s = xml_escape(exp_line.title)
            lin_data.append(Paragraph(s, style_table_data))

            s = ""
            expense = exp_line.parent_expense
            if expense:
                # XXX use if 'spent_for' in schema
                try:
                    wcase = expense.spent_for[0]
                except (IndexError, AttributeError, KeyError):
                    pass
                else:
                    s = xml_escape(wcase.ref)
            lin_data.append(Paragraph(s, style_table_data))

            s = format_number(exp_line.amount, 2)
            lin_data.append(Paragraph(s, style_table_data_num))

            s = xml_escape(exp_line.currency)
            lin_data.append(Paragraph(s, style_table_data))

            s = format_number(exp_line.exchange_rate, 5)
            lin_data.append(Paragraph(s, style_table_data_num))

            s = format_number(exp_line.euro_amount(), 2)
            lin_data.append(Paragraph(s, style_table_data_num))

            tab_data.append(lin_data)

        # Builds the main table containing the expenses lines
        main_table = Table(
            tab_data,
            [2.25 * cm, 7 * cm, 1.75 * cm, 2.25 * cm, 1.25 * cm, 1.75 * cm, 2.75 * cm],
        )
        main_table.repeatRows = 1
        main_table.setStyle(
            TableStyle(
                [
                    ("ALIGN", (0, 0), (-1, -1), "LEFT"),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                    ("LEFTPADDING", (0, 0), (-1, -1), 0.1 * cm),
                    ("RIGHTPADDING", (0, 0), (-1, -1), 0.1 * cm),
                    ("TOPPADDING", (0, 0), (-1, -1), 0.1 * cm),
                    ("BOTTOMPADDING", (0, 0), (-1, -1), 0.1 * cm),
                    ("BOX", (0, 0), (-1, -1), 0.05 * cm, black),
                    ("LINEAFTER", (0, 0), (0, -1), 0.02 * cm, black),
                    ("LINEAFTER", (1, 0), (1, -1), 0.02 * cm, black),
                    ("LINEAFTER", (2, 0), (2, -1), 0.05 * cm, black),
                    ("LINEAFTER", (3, 0), (3, -1), 0.02 * cm, black),
                    ("LINEAFTER", (4, 0), (4, -1), 0.02 * cm, black),
                    ("LINEAFTER", (5, 0), (5, -1), 0.05 * cm, black),
                    ("LINEBELOW", (0, 0), (-1, 0), 0.05 * cm, black),
                ]
            )
        )

        return main_table



"""template automatic tests"""
import unittest
from cubicweb_web.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    def to_test_etypes(self):
        return set(
            ("Expense", "ExpenseLine", "Refund", "PaidByAccount", "PaidForAccount")
        )

    def list_startup_views(self):
        return ()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""cubicweb-expense specific hooks and notification views

:organization: Logilab
:copyright: 2008-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

from itertools import groupby

from logilab.common.textutils import normalize_text

from cubicweb import _
from cubicweb.predicates import is_instance, is_in_state
from cubicweb.server import hook
from cubicweb.sobjects import notification


class UpdateRefundStateOperation(hook.Operation):
    """adds the new expense line automatically to the corresponding
    refund entity. If no refund entity is associated to the account
    who paid the expense, the operation will create one.
    """

    def precommit_event(self):
        cnx = self.cnx
        execute = cnx.execute
        rql = (
            "Any A,U,EL ORDERBY A WHERE X has_lines EL, EL paid_by A, "
            "A associated_to U?, X eid %(x)s"
        )
        acc_rset = execute(rql, {"x": self.expense})
        for account, lines in groupby(acc_rset, lambda row: row[0]):
            # reset refund to None for this new account
            ref = None
            for __, user, line in lines:
                if user is None:  # this is a company account, no refund needed
                    break
                ref = ref or self.get_or_create_refund_for(account)
                # users don't have permissions to add lines to refunds
                execute(
                    "SET R has_lines EL WHERE R is Refund, R eid %(r)s, "
                    "EL eid %(el)s, NOT EXISTS(R has_lines EL)",
                    {"r": ref, "el": line},
                )

    def get_or_create_refund_for(self, account):
        rql = (
            "Any R WHERE R is Refund, R to_account A, A eid %(a)s, "
            'R in_state S, S name "preparation"'
        )
        rset = self.cnx.execute(rql, {"a": account})
        if rset:
            return rset[0][0]
        rql = "INSERT Refund R: R to_account A WHERE A eid %(a)s"
        # users don't have permission to add refunds
        return self.cnx.execute(rql, {"a": account})[0][0]


class OnExpenseAcceptedHook(hook.Hook):
    """
    * when an expense is going from "submitted" to "accepted", create
      (or update) refund entities accordingly
    """

    __regid__ = "expense_accepted"
    __select__ = hook.Hook.__select__ & hook.match_rtype(
        "in_state",
    )
    events = ("after_add_relation",)

    def __call__(self):
        cnx = self._cw
        etype = cnx.entity_type(self.eidfrom)
        if etype != "Expense":
            return
        newstate = cnx.entity_from_eid(self.eidto).name
        if newstate == "accepted":
            UpdateRefundStateOperation(cnx, expense=self.eidfrom)


class ExpenseLinesRecipientsFinder(notification.RecipientsFinder):
    __select__ = is_instance("Expense", "Refund")
    users_rql = (
        "Any X,E,A WHERE EE has_lines EL, EL paid_by AC, AC associated_to X, "
        "X primary_email E, E address A, EE eid %(ee)s"
    )

    def recipients(self):
        expense = self.cw_rset.get_entity(0, 0)
        rset = self._cw.execute(self.users_rql, {"ee": expense.eid})
        return sorted(
            set(
                (
                    u.cw_adapt_to("IEmailable").get_email(),
                    u.property_value("ui.language"),
                )
                for u in rset.entities()
            )
        )


class ExpenseAcceptedView(
    notification.StatusChangeMixIn, notification.NotificationView
):
    __select__ = is_instance("Expense") & is_in_state("accepted")
    content = _(
        """
The expense ``%(title)s`` has been accepted :

Description
-----------
%(description)s

Detail
-----
%(detail)s

URL
---
%(url)s
"""
    )

    def subject(self):
        entity = self.cw_rset.get_entity(0, 0)
        return "%s %s" % (self._cw._("expense accepted: "), entity.title)

    def context(self, **kwargs):
        context = super(ExpenseAcceptedView, self).context(**kwargs)
        entity = self.cw_rset.complete_entity(0, 0)
        description = entity.printable_value("description", format="text/plain")
        description = normalize_text(description, 80)
        detail = "\n".join(line.view("textoutofcontext") for line in entity.has_lines)
        context.update(
            {
                "description": description,
                "title": entity.title,
                "url": entity.absolute_url(),
                "detail": detail,
            }
        )
        return context


class RefundActedView(notification.StatusChangeMixIn, notification.NotificationView):
    __select__ = is_instance("Refund")
    content = _(
        """
Your expenses have been refunded (amount=%(amount)s euros.)

Detail
------
%(detail)s

URL
---
%(url)s
"""
    )

    def subject(self):
        return self._cw._("your expense will be refunded on your next payroll")

    def context(self, **kwargs):
        context = super(RefundActedView, self).context(**kwargs)
        entity = self.cw_rset.complete_entity(0, 0)
        detail = "\n".join(line.view("textoutofcontext") for line in entity.has_lines)
        context.update(
            {
                "amount": entity.total,
                "url": entity.absolute_url(),
                "detail": detail,
            }
        )
        return context



draft = add_state(_("draft"), "Expense", initial=True)
submitted = rql(
    'Any S WHERE S is State, S name "submitted", S state_of E, E name "Expense"'
)[0][0]
add_transition(_("submit"), "Expense", (draft,), submitted)



# expense workflow

from cubicweb import _

wf = add_workflow(_("expense workflow"), "Expense")
draft = wf.add_state(_("draft"), initial=True)
submitted = wf.add_state(_("submitted"))
accepted = wf.add_state(_("accepted"))
wf.add_transition(_("accept"), submitted, accepted, requiredgroups=("managers",))
wf.add_transition(_("submit"), draft, submitted)

# refund workflow
wf = add_workflow(_("refund workflow"), "Refund")
preparation = wf.add_state(_("preparation"), initial=True)
paid = wf.add_state(_("paid"))
pay = wf.add_transition(_("pay"), preparation, paid)



"""base helper classes for eexpense's unittests"""

from datetime import date

from cubicweb_web.devtools.testlib import WebCWTC


class HelpersTC(WebCWTC):
    # helpers
    def add_relation(self, cnx, eidfrom, rtype, eidto):
        cnx.execute(
            "SET X %s Y WHERE X eid %%(x)s, Y eid %%(y)s" % rtype,
            {"x": eidfrom, "y": eidto},
        )

    def new_expense_line(self, cnx, paid_by_eid):
        line = cnx.create_entity(
            "ExpenseLine",
            title="aline",
            diem=date.today(),
            type="food",
            amount=1.0,
            taxes=0.0,
        )
        self.add_relation(cnx, line.eid, "paid_by", paid_by_eid)
        self.add_relation(cnx, line.eid, "paid_for", self.accountfor)
        return line

    def add_expense_line(self, cnx, expense, paid_by_eid):
        line = self.new_expense_line(cnx, paid_by_eid)
        self.add_relation(cnx, expense.eid, "has_lines", line.eid)
        return line.eid

    def accept(self, cnx, expense):
        expense.cw_adapt_to("IWorkflowable").change_state("accepted")
        cnx.commit()  # to fire corresponding operations

    def new_account(self, cnx, login):
        user = self.create_user(cnx, login)
        cnx.execute(
            "INSERT EmailAddress E: E address %(add)s, U use_email E, U primary_email E "
            "WHERE U eid %(u)s",
            {"u": user.eid, "add": login + "@test.org"},
        )
        account = cnx.create_entity("PaidByAccount", label="%s account" % login)
        self.add_relation(cnx, account.eid, "associated_to", user.eid)
        return user, account

    def refund_lines_count(self, cnx, account):
        return cnx.execute(
            "Any COUNT(EL) WHERE R is Refund, R has_lines EL, R to_account A, "
            'A eid %(a)s, R in_state S, S name "preparation"',
            {"a": account},
        )[0][0]

    def create_and_submit_expense(self, cnx):
        expense = cnx.create_entity("Expense", title="company expense")
        self.add_expense_line(cnx, expense, self.account1)
        cnx.commit()
        expense.cw_adapt_to("IWorkflowable").fire_transition("submit")
        return expense

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            add = cnx.create_entity
            # users and accounts initialization
            user1, account1 = self.new_account(cnx, "john")
            user2, account2 = self.new_account(cnx, "bill")
            self.user1, self.user2 = user1.eid, user2.eid
            account_comp = add("PaidByAccount", label="company account")
            accountfor = add("PaidForAccount", label="whatever")
            self.account1, self.account2 = account1.eid, account2.eid
            self.account_comp, self.accountfor = account_comp.eid, accountfor.eid
            # expense creation
            self.expense = add("Expense", title="sprint")
            self.add_expense_line(cnx, self.expense, account1.eid)
            self.add_expense_line(cnx, self.expense, account_comp.eid)
            cnx.commit()
            self.accept(cnx, self.expense)



# -*- coding: iso-8859-15 -*-

"""
Fresh PDF writing library package. This package is based on ReportLab PDF
library, and specifically on platypus (see ReportLab documentation for
details).
"""

# The only class that have to be used for generating PDFs. Just instantiate
# the class:
#     wrt = PDFWriter(company_data)
# and then:
#     wrt.write(entity,output)

from cubicweb_expense.pdfgen.writers import PDFWriter  # noqa: F401



# -*- coding: iso-8859-15 -*-

"""
Library package containing the definition of template classes (page template
and document template) used in the PDF generation with ReportLab platypus.
These classes are derived from platypus standard classes but are specific to
the Fresh application. The page template draw the static content in the
PDF documents.
"""
import six

from reportlab import Version as reportlab_version

from reportlab.lib.units import cm

from reportlab.lib.utils import ImageReader
from reportlab.platypus import PageTemplate, BaseDocTemplate, Frame

from logilab.mtconverter import xml_escape


class FreshPageTemplate(PageTemplate):
    """
    Class defining the page template used to draw the pages of the PDF documents
    generated by Fresh (Expense and Refund). This page template has a main frame
    that will be filled with the entities data, and draws various other elements
    as the logo and the header that are displayed identically on each page
    (static content).
    """

    def __init__(self, doc_type, company_data, template_id="", _=six.text_type):
        """
        Instanciates a FreshPageTemplate.

        doc_type: string. type of the document. Can be "refund"
                  (Refund) or "expense" (Expense).
        company_data: {'':u"", }. dictionnary containing data about the
                      company (logo filename, company address, etc.)
        template_id: string. id of the page template.
        """
        # Defines the main frame in which the expenses data will be displayed
        main_f = Frame(
            1 * cm,
            1 * cm,
            19 * cm,
            23.9 * cm,
            leftPadding=0 * cm,
            bottomPadding=0 * cm,
            rightPadding=0 * cm,
            topPadding=0 * cm,
        )

        self.company_data = company_data
        self.doc_type = doc_type
        self._ = _

        PageTemplate.__init__(self, id=template_id, frames=[main_f])

    def beforeDrawPage(self, canvas, document):
        """
        Overrides the method called before drawing on a new page of the PDF
        document. This method draws various static data on the page
        (logo, company address, header, page number).

        canvas: Canvas. Reportlab canvas that the PDF objects are drawn in.
        document: BaseDocTemplate. Reportlab document template.
        """
        # saves canvas state
        canvas.saveState()

        self.draw_static_content(canvas, document)

        # Puts back the canvas in its previous state
        canvas.restoreState()

    def draw_static_content(self, canvas, document):
        """
        Draws on the PDF page the static content that is displayed on all
        pages, i.e. logo, company address, header, page number.

        canvas: Canvas. Reportlab canvas that the PDF objects are drawn in.
        document: BaseDocTemplate. Reportlab document template.
        """
        _ = self._
        # Here we don't use a Frame and the platypus machinery because we want
        # to draw various elements that are absolutely positioned. Moreover,
        # with platypus, we can't draw an image (logo) that is left-aligned.

        # rectangle around the header
        canvas.setLineWidth(0.05 * cm)
        canvas.rect(1 * cm, 28.7 * cm, 19 * cm, -3.3 * cm, stroke=1, fill=0)

        # company logo
        try:
            logo = ImageReader(self.company_data["logo-filename"])
            if reportlab_version >= "2.1" and reportlab_version <= "2.3":
                canvas.drawImage(
                    logo,
                    1.2 * cm,
                    28.5 * cm,
                    5 * cm,
                    2.5 * cm,
                    preserveAspectRatio=True,
                    anchor="nw",
                )
            else:
                # Old version of reportlab. must compute the height and width
                # of the image. draws the image from the southwest corner.
                # 2.4 Version has the same problem!
                img_width, img_height = logo.getSize()
                ratio = img_width / img_height
                if ratio >= 5 / 2.5:
                    width = 5 * cm
                    height = 5 * cm / ratio
                else:
                    width = ratio * 2.5 * cm
                    height = 2.5 * cm
                canvas.drawImage(logo, 1.2 * cm, 28.5 * cm - height, width, height)
        except IOError:
            # Unable to read logo filename
            canvas.setFont("Helvetica", 12)
            str_data = xml_escape(self.company_data["company-name"])
            canvas.drawString(1.2 * cm, 28 * cm, str_data)

        canvas.setFont("Helvetica", 8)

        # page number
        str_data = _("Page") + _(": ") + "%d" % document.page
        canvas.drawRightString(19.8 * cm, 28.2 * cm, str_data)

        # company address
        str_data = _("Est.") + _(": ") + self.company_data["company-address"]
        self.draw_string_in_width(canvas, str_data, 1.2 * cm, 25.6 * cm, 9.5 * cm)

        # company official ID number
        str_data = _("Official ID num") + _(": ") + self.company_data["company-offnum"]
        self.draw_string_in_width(canvas, str_data, 11.2 * cm, 25.6 * cm, 5 * cm)

        # company activity number
        str_data = _("Activity num") + _(": ") + self.company_data["company-actnum"]
        self.draw_string_in_width(canvas, str_data, 16.7 * cm, 25.6 * cm, 3.5 * cm)

        # title depending on document type
        title = ""
        if self.doc_type == "refund":
            title = _("Refund Document").upper()
        elif self.doc_type == "expense":
            title = _("Expense Document").upper()
        canvas.setFont("Helvetica-Bold", 14)
        canvas.drawCentredString(14.2 * cm, 27.2 * cm, title)

        # preamble remark
        canvas.setFont("Helvetica-Oblique", 9)
        canvas.drawCentredString(
            14.2 * cm,
            26.7 * cm,
            _("All the amounts are displayed in Euros, except if specified"),
        )

    def draw_string_in_width(self, canvas, string, x, y, width):
        """
        In a canvas, draws a string from the x,y position and within the
        specified width. If the value is too large, the string
        is truncated in order to keep it into the specified width. Therefore,
        the string is drawned between the positions (x,y) and (x+width,y).
        This method gently adds "..." at the end of the truncated string.

        canvas: Canvas. Reportalb canvas that the strings are drawn in
        string: unicode string. string to be drawn in the canvas
        x: float. horizontal position that the string is drawn from
        y: float. vertical position that the string is drawn on
        width: float. maximum width of the string.
        """
        if reportlab_version >= "2.1":
            str_width = canvas.stringWidth(string)
        else:
            # Old version of reportlab library. must provide font informations
            str_width = canvas.stringWidth(string, canvas._fontname, canvas._fontsize)
        if str_width > width:
            string = string + "..."
            str_width = canvas.stringWidth(string)
            while str_width > width:
                string = string[:-4] + "..."
                str_width = canvas.stringWidth(string)
        canvas.drawString(x, y, string)


class FreshDocTemplate(BaseDocTemplate):
    """
    The Fresh document template is a class explaining how to create a PDF
    document for the Fresh application (Expense or Refund). It uses
    the FreshPageTemplate.
    """

    def __init__(self, output, doc_type, company_data, _=six.text_type):
        """
        Instanciates a FreshDocTemplate.
        output: string or output flow. name of the PDF output file or
                writable flow where to write the PDF output

        doc_type: string. type of the document that can be "refund"
                  (Refund) or or "expense" (Expense).
        company_data: {'':u"", }. dictionnary containing various data about
                      the company.
        """
        # Initializes the document template with the correct page templates
        templates = [
            FreshPageTemplate(doc_type, company_data, template_id="Pages", _=_)
        ]
        BaseDocTemplate.__init__(
            self, output, pageTemplates=templates, allowSplitting=1
        )

        if doc_type == "refund":
            self.title = _("Refund Document")
        elif doc_type == "expense":
            self.title = _("Expense Document")
        else:
            self.title = ""
        self.author = company_data["company-name"]



# pylint: disable-msg=W0622
"""cubicweb-expense application packaging information"""

modname = "expense"
distname = "cubicweb-expense"

numversion = (1, 1, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "Logilab"
author_email = "contact@logilab.fr"
description = "expense component for the CubicWeb framework"
web = "http://www.cubicweb.org/project/%s" % distname
classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: JavaScript",
]

__depends__ = {
    "cubicweb": ">=4.0.0,<6.0.0",
    "cubicweb-web": ">=1.6.0,<2.0.0",
    "cubicweb-addressbook": ">=2.2.1,<3.0.0",
    "cubicweb-file": ">=4.2.0,<5.0.0",
}



# -*- coding: iso-8859-15 -*-

"""
Library package containing the styles used in the PDF generation with
ReportLab platypus.
"""

from reportlab.lib.units import cm
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT

STANDARD_FONT_SIZE = 9.0

# Defining paragraph styles used in PDF generation

style_metadata = ParagraphStyle("metadata")
style_metadata.fontSize = STANDARD_FONT_SIZE
style_metadata.leading = STANDARD_FONT_SIZE * 1.2
style_metadata.fontName = "Helvetica"
style_metadata.alignment = TA_LEFT
style_metadata.firstLineIndent = -0.5 * cm
style_metadata.leftIndent = 0.5 * cm

style_ids = ParagraphStyle("ids")
style_ids.fontSize = STANDARD_FONT_SIZE * 0.8
style_ids.leading = STANDARD_FONT_SIZE * 0.8 * 1.2
style_ids.fontName = "Helvetica"
style_ids.alignment = TA_LEFT

style_address = ParagraphStyle("address")
style_address.fontSize = STANDARD_FONT_SIZE * 1.2
style_address.leading = STANDARD_FONT_SIZE * 1.2 * 1.2
style_address.fontName = "Helvetica"
style_address.alignment = TA_LEFT
style_address.firstLineIndent = -0.5 * cm
style_address.leftIndent = 0.5 * cm

style_table_header = ParagraphStyle("table_header")
style_table_header.fontSize = STANDARD_FONT_SIZE
style_table_header.leading = STANDARD_FONT_SIZE * 1.1
style_table_header.fontName = "Helvetica"
style_table_header.alignment = TA_CENTER

style_table_data = ParagraphStyle("table_data")
style_table_data.fontSize = STANDARD_FONT_SIZE
style_table_data.leading = STANDARD_FONT_SIZE * 1.1
style_table_data.fontName = "Helvetica"
style_table_data.alignment = TA_LEFT
style_table_data.firstLineIndent = -0.2 * cm
style_table_data.leftIndent = 0.2 * cm

style_table_data_num = ParagraphStyle("table_data_num")
style_table_data_num.fontSize = STANDARD_FONT_SIZE
style_table_data_num.leading = STANDARD_FONT_SIZE * 1.1
style_table_data_num.fontName = "Helvetica"
style_table_data_num.alignment = TA_RIGHT



# -*- coding: iso-8859-15 -*-

"""
Library package containing the definition of useful text formatting
functions.
"""


def format_number(number, digits=2):
    """
    Function that formats a number to be written in the PDF document. It uses
    the separating characters for numbers defined in the localized
    dictionnary

    number: float or int. number to be formatted.
    digits: int. number of digits to be displayed ('number' is rounded if it
            has more digits).
    Returns: unicode string.
    """
    str_mask = "%%.%df" % digits
    # str_num contains the string representation of 'number' rounded with
    # 'digits' digits.
    str_num = str_mask % number

    # gets the integer string
    str_int = str_num.split(".")[0]

    # passes through the integer string (in reverse order) and adds thousands
    # separator when needed
    result_int = ""
    count = 0
    for char in str_int[::-1]:
        if char in "0123456789":
            if count != 0 and count % 3 == 0:
                # result_int = _(u"thousands sep") + result_int
                result_int = "\u00a0%s" % result_int
            result_int = char + result_int
            count += 1
        elif char in "+-":
            result_int = "%s\u00a0%s" % (char, result_int)

    result_int = " ".join(result_int.split())

    if digits == 0:
        formatted_num = result_int
    else:
        formatted_num = "%s,%s" % (result_int, str_num.split(".")[1])
    return formatted_num
